// EUSART(PIC18F) interface routines
//   STX/ETX type comunication
#include <p18f45j10.h>
#include <stdio.h>
#include <stdlib.h>
#include "eusart.h"

unsigned char com_buffer[COMBUFLEN];
unsigned char tx_buffer[COMBUFLEN];
unsigned char com_bufp, com_etx_flag;
int com_timeout;
int tx_len;

// 76.8Kbps, 8bit data, non-Parity
void
iniz_com( void )
{
  unsigned char w;

  TRISDbits.TRISD2 = 0;
  TRISDbits.TRISD3 = 0;
  SPBRG = 51, SPBRGH = 0;
  RCSTA = 0b10010000; // EUSART and continuous receive enable
  BAUDCON = 0b00001000;	// BRG16=1
  TXSTA = 0b00100100;	// transmit enable. BRGH=1
  w = RCREG;	// buffer discard
  com_buffer[0] = tx_buffer[0] = '\0';
  com_bufp = com_etx_flag = 0;
  com_timeout = 0;
  tx_len = 0;
  PORTDbits.RD2 = 0;
  PORTDbits.RD3 = 0;
  PIE1bits.RCIE = 1;	// interrupt enable
  PIE1bits.TXIE = 0;
}

// interrupt service routine - call from main program
void
eusart_isr( void )
{
  unsigned char d;

  if( PIR1bits.RCIF ) {
    d = RCREG;
    PIR1bits.RCIF = 0;
    if( COMBUFLEN > com_bufp ) {
      com_buffer[com_bufp++] = d;
    } else {
      com_etx_flag = COM_BUFFER_FULL;
    }
    if( ETX == d ) com_etx_flag = COM_RECV_ETX;
  }
  if( PIR1bits.TXIF ) {
    PIR1bits.TXIF = 0;
  }
}

unsigned int x, y, z;
// CCITT CRC
unsigned int
crc( unsigned char *d, int len )
{
  unsigned int crc16;
  int i, j;

  crc16 = 0xffff;
  for( i = 0; i < len; i ++ ) {
    crc16 ^= ( ( unsigned int )d[i] ) << 8;
    for( j = 0; j < 8; j ++ ) {
      if( 0x8000 & crc16 ) crc16 = ( crc16 << 1 ) ^ 0x1021;
      else                 crc16 <<= 1;
    }
  }
  return crc16;
}

unsigned char
putc_com( unsigned char c )
{
  while( 0 == TXSTAbits.TRMT );
  TXREG = c;
  return c;
}

int
transmit_com( unsigned char p[], int n )
{
  int i;

  PORTDbits.RD2 = 1;	// RS422 drive enable
  for( i = 0; i < n; i ++ ) putc_com( p[i] );
  while( 0 == TXSTAbits.TRMT );
  PORTDbits.RD2 = 0;	// RS422 drive disable
  return n;
}

// create a transmit packet
//  return code == 0 : error
//              != 0 : packet length
int
create_packet( unsigned int p[], int n )
{
  int i;
  unsigned int c;
  char cs;	// chip select code
  char s[6];

  tx_buffer[0] = '\0';		// first, buffer clear
  tx_len = 0;

  switch( 0x0F00 & p[0] ) {
    case 0x0E00 : cs = '4'; break;
    case 0x0D00 : cs = '3'; break;
    case 0x0B00 : cs = '2'; break;
    case 0x0700 : cs = '1'; break;
    default :  return 0;
  }

  if( 0 < n && COMBUFLEN > n + 5 ) {
    sprintf( &tx_buffer[1], "%02X", 2 * n + 4 );	// length
    tx_buffer[3] = '0';
    tx_buffer[4] = cs;	// instruction
    for( i = 0; i < n; i++ ) {
      sprintf( &tx_buffer[5+i*2], "%02X", p[i] & 0xff );
    }
    c = crc( tx_buffer + 1, n * 2 + 4 );
    sprintf( &tx_buffer[5+n*2], "%04X", c );
    tx_buffer[9+n*2] = ETX;
    tx_buffer[0] = STX;
    return tx_len = 10 + 2 * n;
  }
  else return 0;
}

void
packet_buffer_clear( void )
{
  tx_buffer[0] = '\0';
  tx_len = 0;
  com_timeout = 0;
}

void
recv_buffer_clear( void )
{
  com_buffer[0] = '\0';
  com_bufp = 0;
  com_etx_flag = 0;
}

int
transmit_spi_packet( unsigned int d[], int len )
{
  int i;

  i = create_packet( d, len );
  if( 0 ==i ) return -1;	// error
  com_timeout = TRX_TIMEOUT;
  transmit_com( tx_buffer, tx_len );
  return 0;		// normal return
}

int
retry_packet( void )
{
  if( 0 == tx_len ) return -1;	// no data
  com_timeout = TRX_TIMEOUT;
  transmit_com( tx_buffer, tx_len );
  return 0;		// normal return
}

int
transmit_ACK_packet( void )
{
  unsigned int c;

  tx_buffer[0] = '\0';		// first, buffer clear
  tx_len = 0;
  tx_buffer[1] = 0x01;		// length
  tx_buffer[2] = ACK;		// instruction
  c = crc( tx_buffer + 1, 2 );
  sprintf( &tx_buffer[3], "%04X", c );
  tx_buffer[7] = ETX;
  tx_buffer[0] = STX;
  com_timeout = TRX_TIMEOUT;
  transmit_com( tx_buffer, 8 );
  return 0;		// normal return
}

int
transmit_NAK_packet( void )
{
  unsigned int c;

  tx_buffer[0] = '\0';		// first, buffer clear
  tx_len = 0;
  tx_buffer[1] = 0x01;		// length
  tx_buffer[2] = NAK;		// instruction
  c = crc( tx_buffer + 1, 2 );
  sprintf( &tx_buffer[3], "%04X", c );
  tx_buffer[7] = ETX;
  tx_buffer[0] = STX;
  com_timeout = TRX_TIMEOUT;
  transmit_com( tx_buffer, 8 );
  return 0;		// normal return
}

void
com_buffer_clear( void )
{
  com_buffer[0] = '\0';
  com_bufp = com_etx_flag = com_timeout = 0;
}

unsigned char
hex2bin( char *p )
{
  unsigned char h, l;

  h = *p++; if( '9' < h ) h = h - '7'; else h = h - '0';
  l = *p++; if( '9' < l ) l = l - '7'; else l = l - '0';
  return h * 16 + l;
}

// ACK receive check
//  0: not received
//  1: ACK received
// -1: NAK received or buffer overflow or receive error
int
ack_check( void )
{
  char *d;
  int len;

  if( 0 == com_etx_flag ) return 0;
  if( COM_BUFFER_FULL == com_etx_flag ) {
    com_buffer_clear();
    return -1;
  }
  d = com_buffer;
  len = com_bufp;
  while( STX != *d && 0 != len ) d++, --len;
  d++; // skip STX
  d++; // skip length
  if( ACK == *d ) return 1;	// (ignore CRC)
  return -1;
}


