#include <p18f2520.h>
#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include <ctype.h>
#include <string.h>
#include <eep.h>
#include "i2c_master.h"
#include "st7032.h"

#pragma config OSC=INTIO67, PWRT=OFF, BOREN=OFF, WDT=OFF
#pragma config PBADEN=OFF, MCLRE=ON, LVP=OFF

// Timing control variables 
unsigned int Timer1S = 0, Timer100mS = 0, PS1S = 1;
// USART(COM) buffer, variables
#define COMBUFLEN 64
char ComRxBuf[COMBUFLEN];
int RxLen;
char LastChar;

// measurement settings and variables
#define MAX_POWER 30.0
char meas_curr = 1, meas_volt = 1, meas_temp = 1;
double current_setting = 0.0;
double voltage_meas; // voltage data
int meas_interval = 0;

#define EE_CurrMeasCoef 0
#define EE_CurrMeasOffs sizeof(double)
#define EE_VoltMeasCoef 2*sizeof(double)
#define EE_CurrSetCoef  3*sizeof(double)
#define EE_TempMeasCoef 4*sizeof(double)
#define EE_TempMeasOffs 5*sizeof(double)

// setthing and coefficient value
double CurrentMeasCoefficient = 3.051E-3;
double CurrentMeasOffset = 0.0;
double VoltageMeasCoefficient = 6.714E-3;
double CurrentSetCoefficient = 1164.1;
double TempMeasOffset = 1726.4;
double TempMeasCoefficient = 21.28;

// interrupt handler
void interrupt isr( void )
{
  char w;	// waste buffer

  if( INTCONbits.T0IF ) {
    INTCONbits.T0IF = 0;
    TMR0L = 0xCB; TMR0H = 0xF3;		// reset for 0.1sec.
    if( Timer100mS ) --Timer100mS;	// count down
    if( 0 == --PS1S ) {
      PS1S = 10;
      if( 0 != Timer1S ) --Timer1S;
    }
  }
  if( PIR1bits.RCIF ) {
    PIR1bits.RCIF = 0;
    w = RCREG;		// read data
    if( COMBUFLEN <= RxLen ) return;	// buffer full waste data
     ComRxBuf[RxLen++] = w;
    LastChar = w;
  }
}

void
initialize( void )
{
// CLOCK setthing - 32MHz
  OSCCON = 0x70;		// FRC 8.0MHz
  OSCTUNEbits.PLLEN = 1; // *4 PLL enable ( 32MHz Fosc )
// I/O port direction
  ADCON1 = 0x0f;
  PORTA = 0xF8;
  TRISA = 0b00000010;	// SPI interface - software control
//          ..0.....	MCP3204 CS
//          ...0....	MCP4821 LDAC
//          ....0...	MCP4821 CS
//          .....0..	SCLK
//          ......1.	MISO
//          .......0	MOSI
  TRISB = 0x00;
  TRISC = 0x98;
  PORTA = 0x00;
  PORTB = 0x00;
  PORTC = 0x00;
// interval timer - TMR0(16bit mode)
  TMR0L = 0; TMR0H = 0;
  T0CON = 0b10000111;
//          1.......	TMR1 ON
//          ....0111	1/256 pre-scaler, count up 32uS
  INTCONbits.T0IE = 1;
// CCP1 PWM output - negative power supply for OPAMP
  PR2 = 0xFF;		// period
  CCPR1L = 0x80;	// duty cycle
  T2CON = 0x04;		// TMR2 start
  CCP1CON = 0x0C;	// CCP1 PWM mode
// software I2C
  iniz_i2c();
// ST7032i - LCD driver
  iniz_st7032();
  contrast_st7032( 4 );
// EUSART 9600bps 8bit non-parity
  SPBRG  = 0x40;		// 9600bps
  SPBRGH = 0x03;
  BAUDCON = 0b00001000;		// BRG16 = 1
  RCSTA   = 0b10010000;		// serial port enable, continuous receive enable
  TXSTA   = 0b00100100;		// transmitter enable, BRGH = 1
  RxLen = 0;				// buffer clear
  LastChar = 0;
  ComRxBuf[0] = 0;
  PIE1bits.RCIE = 1;		// serial port receive interrupt enable
// interrupr enable
  INTCONbits.GIE = 1;
}

#define MOSI PORTAbits.RA0
#define MISO PORTAbits.RA1
#define SCLK PORTAbits.RA2
#define DAC  PORTAbits.RA3
#define LDAC PORTAbits.RA4
#define ADC  PORTAbits.RA5

unsigned char
spi8bit( unsigned char d )
{
  int i;
  unsigned char miso;

  for( i = 0, miso = 0; i < 8; i ++ ) {
    if( 0x80 & d ) MOSI = 1;
    else           MOSI = 0;
    SCLK = 1;
    Nop();
    miso <<= 1;
    if( MISO ) miso |= 0x01;
    SCLK = 0;
    d <<= 1;
  }
  return miso;
}

void
mcp4821( int n )
{
  DAC = 0;	// chip select
  spi8bit( 0x30 | ( n / 256 ) );
  spi8bit( n & 0xFF );
  DAC = 1;	// chip select
  LDAC = 0;	// load latch to DAC
  Nop();
  LDAC = 1;
}

int
mcp3204( int ch )
{
  int r;

  ADC = 0;		// chip select
  spi8bit( 0x06 );	// start and single channel
  r = spi8bit( ch * 64 );	// channel
  r = ( r & 0x0f ) * 256;
  r += spi8bit( 0 );
  ADC = 1;
  return r;
}

int
current_set( double cs )
{
  int d, r = 0;

  if( MAX_POWER <= voltage_meas * cs ) {	// POWER LIMIT
    cs = MAX_POWER / voltage_meas;
    r = 1;		// power limit flag
  }
  d = ( unsigned int )( cs * CurrentSetCoefficient );
  if( 4095 < d ) d = 4095, r = 1;
  mcp4821( d );
  return r;
}

char
putc_com( char c )
{
  while( 0 == TXSTAbits.TRMT ) ;
  TXREG = c;
  return c;
}

void
puts_com( char *s )
{
  while( *s ) putc_com( *s++ );
}

double
ee_memory_read( unsigned int addr )
{
  unsigned int i;
  union {
    double d;
    unsigned char c[sizeof(double)];
  } m;

  for( i = 0; i < sizeof(double); i ++ ) {
    m.c[i] = Read_b_eep( addr + i );
  }
  return m.d;
}

void
ee_memory_write( unsigned int addr, double x )
{
  unsigned int i;
  union {
    double d;
    unsigned char c[sizeof(double)];
  } m;

  m.d = x;
  for( i = 0; i < sizeof(double); i ++ ) {
    Write_b_eep( addr + i, m.c[i] );
    Busy_eep();
  }
}

char instruction[6][5] = {
  {"CURR"}, {"MEAS"}, {"INTE"}, {"CONT"}, {"CALI"}
};
char operand[4][5] = {
  {"CURR"}, {"VOLT"}, {"TEMP"}
};

#define TOKEN_LENGTH 16
char token[TOKEN_LENGTH];

char *
get_token( char *p )
{
  int i;

  if( 0 == *p ) return NULL;
  token[0] = '\0';
  i = 0;
  while( i < TOKEN_LENGTH-1 ) {
    // Alphabet - to Upper case
    if( isalpha( *p ) ) {
      while( i < TOKEN_LENGTH-1 ) {
        if( isalpha( *p ) ) token[ i++ ] = toupper( *p++ );
        else {
          token[ i ] = '\0';
          return p;
        }
      }
    }
	// Digit and Dot
    else if( isdigit( *p ) ) {
      while( i < TOKEN_LENGTH-1 ) {
        if( isdigit( *p ) || '.' == *p ) token[ i++ ] = *p++;
        else {
          token[ i ] = '\0';
          return p;
       }
      }
    }
	// separator
    else if( ',' == *p || ':' == *p ) {
      token[ i ++ ] = *p++;
      token[ i ] = '\0';
      return p;
    }
	// skip space and line feed
    else if( ' ' == *p || '\t' == *p || 0x0a == *p ) {
      p++;
      continue;
    }
    else return NULL;
  }
  return NULL;
}

void
clear_ComRxBuf( void )
{
  ComRxBuf[0] = '\0';
  RxLen = 0;
  LastChar = '\0';
}

double
a2f( char *p )
{
  double x, d;
  int i;

  x = 0.0;
  d = 1.0;

  for( i = 0; i < 10; i ++ ) {
    if( isdigit( *p ) ) {
      x = x * 10.0 + ( double )( *p - '0' );
      p++;
    }
    else if( '.' == *p ) break;
    else return x;
  }
  if( '.' == *p ) {
    p++;
    for( i = 0; i < 10; i ++ ) {
      if( isdigit( *p ) ) {
        d = d / 10.0;
        if( 0.0 == d ) break;
        x = x + d * ( double )( *p - '0' );
        p++;
      }
      else break;
    }
  }
  return x; 
}

void
command_interpreter( void )
{
  int i;
  char s[16];
  char *p;
  double x;

  p = ComRxBuf;
  while( p ) {
    p = get_token( p );
    token[7] = '\0';
  // CURRent set
    if( 0 == strcmp( token, instruction[0] ) ) {
      p = get_token( p );
      if( isdigit( token[0] ) ) {
        current_setting = a2f( token );
        i = current_set( current_setting );
        sprintf( s, "%+2.3f,", current_setting );
        command_st7032( 0xC2 );
        puts_st7032( s );
        command_st7032( 0xC0 );
        if( i ) data_st7032( 'L' );
        else    data_st7032( ' ' );
        clear_ComRxBuf();
        return;
      }
      else {
        sprintf( s, "ERR %s\r\n", token );
        puts_com( s );
        clear_ComRxBuf();
        return;
      }
    }
 // MEASure
    else if( 0 == strcmp( token, instruction[1] ) ) {
      meas_curr = meas_volt = meas_temp = 0;
      while( 1 ) {
        p = get_token( p );
        if( NULL == p ) {
	      clear_ComRxBuf();
          meas_curr = 1; meas_volt = 1; meas_temp = 1;
          return;
        }
        if( ':' == token[0] || ',' == token[0] ) continue;	// skip separator
        if( 0 == strcmp( token, operand[0] ) ) meas_curr = 1;		// current
        else if( 0 == strcmp( token, operand[1] ) ) meas_volt = 1;	// voltage
        else if( 0 == strcmp( token, operand[2] ) ) meas_temp = 1;	// temparature
        else {
	      clear_ComRxBuf();
          return;
        }
      }
    }
  // INTErval time
    else if( 0 == strcmp( token, instruction[2] ) ) {
      p = get_token( p );
      if( NULL == p ) {
        clear_ComRxBuf();
        meas_interval = 0;
        return;
      }
      meas_interval = atoi( token );
      if( 0 > meas_interval ) meas_interval = 0;
      if( 600 < meas_interval ) meas_interval = 600;
      Timer1S = 1;
      clear_ComRxBuf();
      return;
    }
  // CONTrast LCD
    else if( 0 == strcmp( token, instruction[3] ) ) {
      p = get_token( p );
      if( NULL == p ) {
        clear_ComRxBuf();
        meas_interval = 0;
        return;
      }
      i = atoi( token );
      if( 0 > i ) i = 0;
      if( 31 < i ) i = 31;
      contrast_st7032( i );
      clear_ComRxBuf();
      return;
    }
  // CALIbration
    else if( 0 == strcmp( token, instruction[4] ) ) {
      p = get_token( p );
      if( NULL == p ) {
        for( i = 0; 0xFF != Read_b_eep( i * sizeof(double) ); i ++ ) {
          x = ee_memory_read( i * sizeof(double) );
          sprintf( s, "%3d,", i );
          puts_com( s );
          sprintf( s, "%f\r\n", x );
          puts_com( s );
        }
        clear_ComRxBuf();
        return;
      }
      i = atoi( token );	// # of calibration data
      p = get_token( p );
      x = a2f( token );		// value
      ee_memory_write( i * sizeof(double), x );
      sprintf( s, "%3d,", i );
      puts_com( s );
      sprintf( s, "%f\r\n", x );
      puts_com( s );
      clear_ComRxBuf();
      return;
    }
    else {
      sprintf( s, "ERR %s\r\n", token );
      puts_com( s );
      clear_ComRxBuf();
      return;
    }
  }
}

int
main()
{
  int i, j, k, a, ct;
  long int c, v, t;
  char s[16];
  double x;

  initialize();
  contrast_st7032( 0x18 );

// start message
    sprintf( s, "USB" );
    command_st7032( 0x80 );
    puts_st7032( s );
    sprintf( s, "DC ELECTRIC LOAD" );
    command_st7032( 0xC0 );
    puts_st7032( s );
// calinration data read
  if( 0xff != Read_b_eep( 0 ) ) {
    CurrentMeasCoefficient = ee_memory_read( EE_CurrMeasCoef );
    CurrentMeasOffset = ee_memory_read( EE_CurrMeasOffs );
    VoltageMeasCoefficient = ee_memory_read( EE_VoltMeasCoef );
    CurrentSetCoefficient = ee_memory_read( EE_CurrSetCoef );
    TempMeasCoefficient = ee_memory_read( EE_TempMeasCoef );
    TempMeasOffset = ee_memory_read( EE_TempMeasOffs );
  }
// current measure offset calib.
  mcp4821( 0 );
  c = 0; a = 0;
  for( Timer100mS = 10; 0 != Timer100mS ; a++ ) {	
    c += ( long )mcp3204( 0 );
  }
  c = c / ( long )a;
  CurrentMeasOffset = 5.0 - ( double )c * CurrentMeasCoefficient;
  command_st7032( 0x01 );	// display clear
// main loop
  for( ; ; ) {
    if( 0x0d == LastChar ) command_interpreter();
    Timer100mS = 10;	// 1.0 sec interval
    c = v = t = 0;
    for( a = 0; Timer100mS; a ++ ) {		// average count
      c += ( long )mcp3204( 0 );
      v += ( long )mcp3204( 1 );
      t += ( long )mcp3204( 2 );
    }
    // interval measure timing control
    if( 0 == Timer1S && 0 != meas_interval ) {
      ct = 1;
      Timer1S = meas_interval;
    }
    // display current
    c = c / ( long )a;
    x = 5.0 - ( double )c * CurrentMeasCoefficient - CurrentMeasOffset;
    sprintf( s, "%+2.2fA ", x );
    command_st7032( 0x82 );
    puts_st7032( s );
    if( meas_curr && ct ) {
      sprintf( s, "%+2.3f", x );
      puts_com( s );
    }
    // display voltage
    v = v / ( long )a;
    x = ( double )v * VoltageMeasCoefficient;
    voltage_meas = x;
    sprintf( s, "%2.2fV", x );
    command_st7032( 0x8A );
    puts_st7032( s );
    if( meas_volt && ct ) {
      sprintf( s, ",%+2.3f", x );
      puts_com( s );
    }
    // display 104JT thermister
    x = ( double )t / ( double )a;
    x = ( x - TempMeasOffset ) / - TempMeasCoefficient;
    sprintf( s, "%3.1f\xdf", x );
    command_st7032( 0xCA );
    puts_st7032( s );
    data_st7032( 'C' );
    data_st7032( ' ' );
    if( meas_temp && ct ) {
      sprintf( s, ",%2.1f", x );
      puts_com( s );
    }
    if( ct && ( meas_curr || meas_volt || meas_temp ) ) {
      putc_com( 0x0d );
      putc_com( 0x0a );
    }
    ct = 0;
  }
}
